; -----------------------------------------------------------------------------
; "SUPER INPUT Modern" by Einar Saukas
; -----------------------------------------------------------------------------

VARIABLE_NAME   EQU  'z'

PO_STORE        EQU   2780
PO_FETCH        EQU   2819
LOOK_VARS       EQU  10418
LASTK           EQU  23560
CH_ADD          EQU  23645

; -----------------------------------------------------------------------------
; PART 1: LOCATE STRING
; -----------------------------------------------------------------------------
; This routine searches for the 1-dimension string array variable z$(), that
; will be used as input buffer in BASIC.

        org     65000

main:
        ld      hl, (CH_ADD)
        push    hl
        ld      hl, VARSTR
        ld      (CH_ADD), hl
        call    LOOK_VARS               ; locate string array z$()
        pop     de
        ld      (CH_ADD), de
        ret     c                       ; variable not found?
        inc     hl

; String array z$() should have 1 dimension only
validate:
        inc     hl
        inc     hl
        ld      a, (hl)                 ; number of dimensions
        dec     a
        ret     nz                      ; 1 dimension?
        inc     hl

; -----------------------------------------------------------------------------
; PART 2: INPUT STRING
; -----------------------------------------------------------------------------
; Assembly programs can skip the previous part and call "main" directly,
; providing in HL the input buffer address. To support input strings with
; at most NN characters, the input buffer should have NN+2 bytes as follows:
;
;   NNccccccccccccccccccccccccc
;
; The first 2 bytes must contain value NN, and the following NN bytes must
; contain the initial string value that will appear in the input area.
;
; Technically, this routine implementation keeps several "pointers" to the
; input buffer, represented as follows:
;
;   pBUFFER    CURSOR
;   |             |
;   NNccccccccccccccccccccccccc
;    |           |            |
;    pSTART      pCURRENT     pLAST
;
; The entire routine is optimized by code size (there would be no advantage
; to optimize it by speed since we need to wait for pressed keys anyway)
; and the data above is stored as follows:
;
; * pBUFFER is typically stored at address BUFFER
; * pCURRENT is typically stored in register pair DE
; * pLAST is typically stored in the stack
; * the cursor character is typically stored in register A'

main_input:
        res     3, (iy+$30)             ; disable CAPS
        ld      a, $a5
set_cursor:
        xor     $2d                     ; convert $a5 -> $88 or $0d -> $20
        ld      (CURSOR), a             ; initialize cursor
        ld      (BUFFER),hl             ; hl = pBUFFER
        ld      c, (hl)
        inc     hl
        ld      b, (hl)                 ; hl = pSTART; bc = NN
        add     hl, bc                  ; hl = pLAST; bc = NN
        push    hl                      ; keep pLAST in the stack
find_last:
        ld      a, 32                   ; put cursor after the last non-space character
        cpd
        jr      nz, buffer_full
find_loop:
        cpd
        jr      z, find_loop
        inc     hl
buffer_full:
        ex      de, hl                  ; de = pCURRENT

print_string:
        call    PO_FETCH                ; preserve PRINT AT position
        push    hl
        push    bc
        ld      hl, (BUFFER)            ; hl = pBUFFER; de = pCURRENT
        ld      c, (hl)
        inc     hl
        ld      b, (hl)                 ; hl = pSTART; de = pCURRENT; bc = NN
print_next:
        inc     hl
        ld      a, (hl)                 ; print next character
        rst     $10
        push    hl
        scf
        sbc     hl, de                  ; at cursor position?
        jr      nz, no_cursor
        ld      hl, SCURSOR             ; if so, print cursor
over_loop:
        ld      a, (hl)
        rst     $10
        xor     a
        cp      (hl)
        inc     hl
        jr      nz, over_loop
no_cursor:
        pop     hl
        dec     bc
        ld      a, b
        or      c
        jr      nz, print_next
print_end:
        pop     bc
        pop     hl
        call    PO_STORE                ; restore PRINT AT position

reset_key:
        xor     a                       ; erase previously processed key
        ld      (LASTK), a

read_key:                               ; read next key, with support for "extended" characters
        ld      a, (LASTK)
        ld      hl, conversion_table
        ld      bc, 15                  ; de = pCURRENT; bc = 15
        cpir
        jr      nz, chk_exit
        ld      a, (hl)                 ; de = pCURRENT; bc < 15

chk_exit:
        pop     hl                      ; hl = pLAST; de = pCURRENT
        ex      af, af'
        ld      a, (CURSOR)
        cp      32
        jr      nz, chk_shift_2
        pop     hl                      ; hl = pLAST; de = pCURRENT
        ld      hl, (BUFFER)            ; hl = pBUFFER; de = pCURRENT
        ex      de, hl                  ; hl = pCURRENT; de = pBUFFER
        inc     hl
        cp      (hl)
        dec     hl
        ccf
        sbc     hl, de                  ; hl = LEN; de = pBUFFER
        ld      b, h
        ld      c, l                    ; hl = LEN; de = pBUFFER; bc = LEN
        ret

chk_shift_2:
        push    hl                      ; hl = pLAST; de = pCURRENT
        ld      c, a
        ex      af, af'
        cp      6
        jr      nz, chk_shift_3
        ld      a, (iy+$30)
        xor     $08
        ld      (iy+$30), a
        ld      a, 2
switch_cursor:
        xor     c
        or      1
        cp      $8f
        sbc     a, b
        ld      (CURSOR), a
print_string2:
        jr      print_string

chk_shift_3:
        cp      4
        jr      z, switch_cursor

chk_chars:
        cp      32
        jr      c, chk_shift_4
        cp      128
        jr      nc, chk_shift_4
        sbc     hl, de                  ; hl = pLAST-pCURRENT-1; de = pCURRENT
        ex      af, af'
        bit     2, a
        jr      nz, overwrite_mode
insert_mode:
        inc     hl                      ; hl = pLAST-pCURRENT; de = pCURRENT
        ld      b, h
        ld      c, l                    ; hl = pLAST-pCURRENT; de = pCURRENT; bc = pLAST-pCURRENT
        add     hl, de                  ; hl = pLAST; de = pCURRENT; bc = pLAST-pCURRENT
        ld      d, h
        ld      e, l                    ; hl = pLAST; de = pLAST; bc = pLAST-pCURRENT
        dec     hl                      ; hl = pLAST-1; de = pLAST; bc = pLAST-pCURRENT
        lddr                            ; hl = pCURRENT-1; de = pCURRENT; bc = 0
overwrite_mode:
        ex      af, af'
        inc     de                      ; de = pCURRENT+1
        ld      (de), a
        jr      nz, print_string2       ; buffer not full
        dec     de                      ; de = pCURRENT
        jr      print_string2

chk_shift_4:
        cp      5
find_last2:
        jp      z, find_last

chk_enter:
        ld      hl, (BUFFER)            ; hl = pBUFFER; de = pCURRENT
        cp      13
        jp      z, set_cursor

chk_shift_1:
        cp      7
        call    z, main_clear           ; hl = pLAST; de = pCURRENT; bc = 0
        jr      z, find_last2

chk_shift_5:
        inc     hl
        ld      c, 1                    ; hl = pSTART; de = pCURRENT; bc = 1
        cp      8
        jr      nz, chk_shift_8
move_left:
        ex      de, hl                  ; hl = pCURRENT; de = pSTART
        sbc     hl, bc                  ; hl = pCURRENT'; de = pSTART
        sbc     hl, de                  ; hl = pCURRENT'-pSTART; de = pSTART
move_check:
        jr      c, print_string2
        add     hl, de                  ; hl = pCURRENT'
        ex      de, hl
print_string3:
        jr      print_string2

chk_shift_8
        cp      9
        jr      nz, chk_shift_6
move_right:
        ex      de, hl                  ; hl = pCURRENT; de = pSTART
        add     hl, bc                  ; hl = pCURRENT'; de = pSTART
        pop     de                      ; hl = pCURRENT'; de = pLAST
        push    de
        dec     de                      ; hl = pCURRENT'; de = pLAST-1
        sbc     hl, de                  ; hl = pCURRENT'-pLAST+1; de = pLAST-1
        ccf
        jr      move_check

chk_shift_6:
        ld      c, 32                   ; hl = pSTART; de = pCURRENT; bc = 32
        cp      10
        jr      z, move_right

chk_shift_7:
        cp      11
        jr      z, move_left

chk_shift_0:                            ; hl = pSTART; de = pCURRENT
        cp      12
        jr      nz, chk_shift_9
        sbc     hl, de                  ; hl = pSTART-pCURRENT; de = pCURRENT
        jr      z, print_string2
        dec     de                      ; hl = pSTART-pCURRENT; de = pCURRENT-1
        jr      delete

chk_shift_9:
        cp      15
        jp      nz, read_key
delete:
        pop     hl                      ; hl = pLAST; de = pCURRENT'
        push    hl                      ; hl = pLAST; de = pCURRENT'
        and     a
        sbc     hl, de                  ; hl = pLAST-pCURRENT'; de = pCURRENT'
        ld      b, h
        ld      c, l                    ; hl = pLAST-pCURRENT'; de = pCURRENT'; bc = pLAST-pCURRENT'
        push    de                      ; hl = pLAST-pCURRENT'; de = pCURRENT'
        inc     de                      ; hl = pLAST-pCURRENT'; de = pCURRENT'+1; bc = pLAST-pCURRENT'
        ld      h, d
        ld      l, e                    ; hl = pCURRENT'+1; de = pCURRENT'+1; bc = pLAST-pCURRENT'
        inc     hl                      ; hl = pCURRENT'+2; de = pCURRENT'+1; bc = pLAST-pCURRENT'
        ldir                            ; hl = pLAST+2; de = pLAST+1; bc = 0
        dec     hl
        dec     hl                      ; hl = pLAST; de = pLAST+1; bc = 0
        ld      (hl), 32
        pop     de                      ; hl = pLAST; de = pCURRENT'; bc = 0
        jr      print_string3

; -----------------------------------------------------------------------------
; PART 3: CLEAR STRING
; -----------------------------------------------------------------------------
; This sub-routine fills the input buffer with spaces and it's automatically
; called when the user presses "Shift-1".
;
; However, Assembly programs can also call it directly to "clear" an input
; buffer (providing in HL the input buffer address) before calling the main
; routine.

main_clear:
        ld      c, (hl)
        inc     hl
        ld      b, (hl)
clear_next:
        inc     hl
        ld      (hl), 32
        dec     bc
        ld      a, b
        or      c
        jr      nz, clear_next
        ret

; -----------------------------------------------------------------------------
; PART 4: DATA AREA
; -----------------------------------------------------------------------------

conversion_table:
        defb    $c3, '|'
        defb    $c5, ']'
        defb    $c6, '['
        defb    $cb, '}'
        defb    $cc, '{'
        defb    $cd, '\'
        defb    $e2, '~'
        defb    $ac, $7f

BUFFER  defw    0

VARSTR  defb    VARIABLE_NAME, '$', ':'

SCURSOR defb    8, 21, 1
CURSOR  defb    0, 21, 0

; -----------------------------------------------------------------------------
